/*
  +----------------------------------------------------------------------+
  | PHP Version 4                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2004 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3 of the PHP license,         |
  | that is bundled with this package in the file LICENSE, and is        |
  | available at through the world-wide-web at                           |
  | http://www.php.net/license/3.txt.                                    |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Authors:    Marco Tabini <marcot@tabini.ca>                          |
  |             Ilia Alshanetsky <ilia@php.net>                          |
  +----------------------------------------------------------------------+

  $Id: pop3.c,v 1.12 2004/01/10 21:27:45 wez Exp $ 
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_pop3.h"

#include <sys/types.h>
#ifndef PHP_WIN32
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#endif
#include <libspopc.h>

/* {{{ Internal macros */

#define POP3_VERSION "0.1"
#define POP3_RESOURCE_NAME	"POP3 Library Session"

#define POP3_SAFE_POINTER(ptr)	\
	if (!ptr) {	\
		php_error_docref(NULL TSRMLS_CC, E_ERROR, "POP3: Memory error"); \
	}

/* libpop3 errors come directly from the pop3 server and
 * contain terminating \r\n after each error string 
 */
#define POP3_CLEAN_ERROR(str) \
    { \
    	char *e = str + strlen(str) - 1; \
    	while (e > str && (*e == '\n' || *e == '\r')) { \
    		e--; \
    	} \
    	*(e + 1) = '\0'; \
    }

#define POP3_CALL(function_call)	\
	pop3_return_val = function_call;	\
	if (pop3_return_val) {	\
		if (pop3_error(pop3_return_val)) {	\
			pop3_disconnect(*socket_id);	\
			POP3_CLEAN_ERROR(pop3_return_val) \
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "%s", pop3_return_val);	\
			free(pop3_return_val);	\
			RETURN_FALSE;	\
		}	\
	} else {	\
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Connection failed");	\
		RETURN_FALSE;	\
	}

#define POP3_CALL_DESTROY(function_call)	\
	POP3_CALL(function_call);	\
	free(pop3_return_val);

#define POP3_FETCH_SESSION(session, session_container)	\
	ZEND_FETCH_RESOURCE (session, int *, &session_container, -1, POP3_RESOURCE_NAME, le_pop3_resource);

#ifdef ZEND_ENGINE_2

#define POP3_REGISTER_OBJECT(_object, _ptr) \
    { \
        pop3_object *obj; \
        obj = (pop3_object*)zend_object_store_get_object(_object TSRMLS_CC); \
        obj->ptr = _ptr; \
    }

#define POP3_FROM_OBJECT(socket_id, object) \
    { \
	pop3_object *obj = zend_object_store_get_object(object TSRMLS_CC); \
	socket_id = obj->ptr; \
	if (!socket_id) { \
        	php_error_docref(NULL TSRMLS_CC, E_WARNING, "The pop3 connection was not established."); \
		RETURN_FALSE; \
	} \
    }
        
#define POP3_DECLARE_INIT_OBJECT(object) \
	zval	*object = getThis();

#else /* ZEND_ENGINE_2 */

#define POP3_REGISTER_OBJECT(_object, _ptr) {}
#define POP3_FROM_OBJECT(socket_id, object) {}

#define POP3_DECLARE_INIT_OBJECT(object)
#define object 0

#endif /* ZEND_ENGINE_2 */


/* }}} */


#define PHP_POP3_HEADER		1
#define PHP_POP3_MESSAGE	2
#define PHP_POP3_DELETE		3

/* True global resources - no need for thread safety here */

static int le_pop3_resource;

/* }}} */

#ifdef ZEND_ENGINE_2

static zend_object_handlers pop3_object_handlers;

/* {{{ pop3_objects_dtor
 */
static void pop3_objects_dtor(void *object, zend_object_handle handle TSRMLS_DC)
{
	pop3_object *intern = (pop3_object *) object;

	if (intern->zo.properties) {
		zend_hash_destroy(intern->zo.properties);
		FREE_HASHTABLE(intern->zo.properties);
	}

	if (intern->ptr)
	{
		char *pop3_return_val = pop3_quit(*((int *)intern->ptr));
		if (pop3_return_val) {
			free(pop3_return_val);
		}
		efree(intern->ptr);
	}

	efree(intern);
}
/* }}} */

/* {{{ pop3_objects_new
 */
PHP_POP3_API zend_object_value pop3_objects_new(zend_class_entry *class_type TSRMLS_DC)
{
	zend_object_value retval;
	pop3_object *intern;
	zval *tmp;

	intern = emalloc(sizeof(pop3_object));
	intern->zo.ce = class_type;
	intern->zo.in_get = 0;
	intern->zo.in_set = 0;
	intern->zo.properties = NULL;
	intern->ptr = NULL;

	ALLOC_HASHTABLE(intern->zo.properties);
	zend_hash_init(intern->zo.properties, 0, NULL, ZVAL_PTR_DTOR, 0);
	zend_hash_copy(intern->zo.properties, &class_type->default_properties, (copy_ctor_func_t) zval_add_ref, (void *) &tmp, sizeof(zval *));

	retval.handle = zend_objects_store_put(intern, pop3_objects_dtor, NULL TSRMLS_CC);
	retval.handlers = (zend_object_handlers *) &pop3_object_handlers;

	return retval;
}
/* }}} */

#endif /* ZEND_ENGINE_2 */

/* {{{ Forward declarations */

/*
 * Forward declarations
 */

void pop3_resource_destructor (zend_rsrc_list_entry *rsrc TSRMLS_DC);

/* }}} */

/* {{{ pop3_functions[]
 *
 * Every user visible function must have an entry in pop3_functions[].
 */
function_entry pop3_functions[] = {
	PHP_FE(pop3_open,	NULL)
	PHP_FE(pop3_get_message_count, NULL)
	PHP_FE(pop3_get_account_size, NULL)
	PHP_FE(pop3_get_message_list, NULL)
	PHP_FE(pop3_get_message_header, NULL)
	PHP_FE(pop3_get_message, NULL)
	PHP_FE(pop3_delete_message, NULL)
	PHP_FE(pop3_undelete, NULL)
	{NULL, NULL, NULL}
};
/* }}} */

#define POP3_ME_MAPPING(name, func_name, arg_types) \
        ZEND_NAMED_FE(name, ZEND_FN(func_name), arg_types)

/* {{{ pop3_class_functions
 */
function_entry pop3_class_functions[] = {
	POP3_ME_MAPPING(pop3, pop3_open, NULL)
	POP3_ME_MAPPING(get_message_count, pop3_get_message_count, NULL)
	POP3_ME_MAPPING(get_account_size, pop3_get_account_size, NULL)
	POP3_ME_MAPPING(get_message_list, pop3_get_message_list, NULL)
	POP3_ME_MAPPING(get_message_header, pop3_get_message_header, NULL)
	POP3_ME_MAPPING(get_message, pop3_get_message, NULL)
	POP3_ME_MAPPING(delete_message, pop3_delete_message, NULL)
	POP3_ME_MAPPING(undelete, pop3_undelete, NULL)
	{NULL, NULL, NULL}
};
/* }}} */


/* {{{ pop3_module_entry
 */
zend_module_entry pop3_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	"pop3",
	pop3_functions,
	PHP_MINIT(pop3),
	PHP_MSHUTDOWN(pop3),
	PHP_RINIT(pop3),		/* Replace with NULL if there's nothing to do at request start */
	PHP_RSHUTDOWN(pop3),	/* Replace with NULL if there's nothing to do at request end */
	PHP_MINFO(pop3),
#if ZEND_MODULE_API_NO >= 20010901
	POP3_VERSION,
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_POP3
ZEND_GET_MODULE(pop3)
#endif

/* {{{ PHP_INI
 */
/* Remove comments and fill if you need to have entries in php.ini
PHP_INI_BEGIN()
    STD_PHP_INI_ENTRY("pop3.global_value",      "42", PHP_INI_ALL, OnUpdateInt, global_value, zend_pop3_globals, pop3_globals)
    STD_PHP_INI_ENTRY("pop3.global_string", "foobar", PHP_INI_ALL, OnUpdateString, global_string, zend_pop3_globals, pop3_globals)
PHP_INI_END()
*/
/* }}} */

/* {{{ php_pop3_init_globals
 */
/* Uncomment this function if you have INI entries
static void php_pop3_init_globals(zend_pop3_globals *pop3_globals)
{
	pop3_globals->global_value = 0;
	pop3_globals->global_string = NULL;
}
*/
/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(pop3)
{
#ifdef ZEND_ENGINE_2

	zend_class_entry _pop3_class_entry;
	INIT_CLASS_ENTRY(_pop3_class_entry, "pop3", pop3_class_functions);
	_pop3_class_entry.create_object = pop3_objects_new;
	pop3_class_entry = zend_register_internal_class(&_pop3_class_entry TSRMLS_CC);

	/* copy the standard object handlers to you handler table */
	memcpy(&pop3_object_handlers, zend_get_std_object_handlers(), sizeof(zend_object_handlers));

#endif /* ZEND_ENGINE_2 */

	/* If you have INI entries, uncomment these lines 
	ZEND_INIT_MODULE_GLOBALS(pop3, php_pop3_init_globals, NULL);
	REGISTER_INI_ENTRIES();
	*/

	le_pop3_resource = zend_register_list_destructors_ex (pop3_resource_destructor, NULL, POP3_RESOURCE_NAME, module_number);

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(pop3)
{
	/* uncomment this line if you have INI entries
	UNREGISTER_INI_ENTRIES();
	*/
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_RINIT_FUNCTION
 */
PHP_RINIT_FUNCTION(pop3)
{
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_RSHUTDOWN_FUNCTION
 */
PHP_RSHUTDOWN_FUNCTION(pop3)
{
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(pop3)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "pop3 support", "enabled");
	php_info_print_table_header(2, "Extension version", POP3_VERSION " - $id$");
	php_info_print_table_end();

	/* Remove comments if you have entries in php.ini
	DISPLAY_INI_ENTRIES();
	*/
}
/* }}} */

/* {{{ pop3_resource_destructor
 */

void pop3_resource_destructor (zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
	if (rsrc->ptr) {
		char *pop3_return_val = pop3_quit(*((int *)rsrc->ptr));
		if (pop3_return_val) {
			free(pop3_return_val);
		}
		efree(rsrc->ptr);
		rsrc->ptr = NULL;
	}
}

/* }}} */
		
/* {{{ proto resource pop3_open(string server_name, string user_name, string password, [int port_number])
  */
PHP_FUNCTION(pop3_open)
{
	char			*server_name, *user_name, *password;
	int			server_name_length, user_name_length, password_length;
	long			port_number = 110;

	int			socket_id[1], *sock_id;

	struct hostent		server;
	struct sockaddr_in	connection;

	char			*pop3_return_val;
	POP3_DECLARE_INIT_OBJECT (object)

	/* Parse parameters */

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sss|l", &server_name, &server_name_length, &user_name,
				&user_name_length, &password, &password_length, &port_number) == FAILURE)
	{
		RETURN_FALSE;
	}

	/* Attempt to initialize the session */
	*socket_id = pop3_prepare(server_name, port_number, &connection, &server);

	if (*socket_id == -1) {
		efree(socket_id);
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unable to connect to server");
		RETURN_FALSE;
	}

	/* Attempt to connect */
	POP3_CALL_DESTROY(pop3_connect(*socket_id, &connection))

	/* Send username */
	POP3_CALL_DESTROY(pop3_user(*socket_id, user_name))
	
	/* Send password */
	POP3_CALL_DESTROY(pop3_pass(*socket_id, password))

	sock_id = emalloc(sizeof(int));
	*sock_id = *socket_id;

	/* Create and return resource or object */
	if (object) {
		POP3_REGISTER_OBJECT(object, sock_id);
	} else {
		ZEND_REGISTER_RESOURCE(return_value, sock_id, le_pop3_resource);
	}
}
/* }}} */

/* {{{ _pop3_get_stats() */

static void _pop3_get_stats(INTERNAL_FUNCTION_PARAMETERS, int mode)
{
	POP3_DECLARE_INIT_OBJECT(object)
	zval	*session_container;
	char	*pop3_return_val;
	int		*socket_id;

	/* Retrieve session data */
	if (object) {
		POP3_FROM_OBJECT(socket_id, object);
	} else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &session_container) == FAILURE) {
			RETURN_FALSE;
		}
		POP3_FETCH_SESSION(socket_id, session_container);
	}

	POP3_CALL(pop3_stat(*socket_id));

	if (!mode) {
		/* Stat number of messages */
		RETVAL_LONG(stat2last(pop3_return_val));
	} else {
		/* return pop3 box size */
		RETVAL_LONG(stat2bytes(pop3_return_val));
	}

	free(pop3_return_val);
}

/* }}} */

/* {{{ proto long pop3_get_message_count (resource session)
  */
PHP_FUNCTION(pop3_get_message_count)
{
	_pop3_get_stats(INTERNAL_FUNCTION_PARAM_PASSTHRU, 0);
}
/* }}} */

/* {{{ proto long pop3_get_account_size (resource session)
  */
PHP_FUNCTION(pop3_get_account_size)
{
	_pop3_get_stats(INTERNAL_FUNCTION_PARAM_PASSTHRU, 1);
}
/* }}} */

/* {{{ proto array pop3_get_message_list (resource session)
  */
PHP_FUNCTION(pop3_get_message_list)
{
	POP3_DECLARE_INIT_OBJECT (object)
	zval	*session_container;
	char	*pop3_return_val;
	int		*socket_id;
	int		*size_list;
	char	**uidl_list;
	int		message_count, i;

	/* Retrieve session data */
	if (object) {
		POP3_FROM_OBJECT(socket_id, object);
	} else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &session_container) == FAILURE) {
			RETURN_FALSE;
		}
		POP3_FETCH_SESSION(socket_id, session_container);
	}

	/* Retrieve message size list */
	POP3_CALL(pop3_list(*socket_id, 0));

	size_list = list2array(pop3_return_val);

	POP3_SAFE_POINTER(size_list);

	free(pop3_return_val);

	/* Retrieve message IDs */

	POP3_CALL(pop3_uidl(*socket_id, 0));

	uidl_list = uidl2array(pop3_return_val);

	POP3_SAFE_POINTER(uidl_list);

	free(pop3_return_val);

	/* Create return array */
	array_init(return_value);

	/* Cycle through messages */
	message_count = size_list[0];

	for (i = 1; i <= message_count; i++) {
		zval *row;
	
		/* Create row array */
		MAKE_STD_ZVAL(row);
		array_init(row);

		add_assoc_long(row, "message_id", i);
		add_assoc_long(row, "size", size_list[i]);
		add_assoc_string(row, "uid", uidl_list[i], 1);

		/* Insert row in return value */

		add_index_zval(return_value, i, row);
	}

	/* Clean up */
	free(size_list);
	freeuidlarray(uidl_list);
}
/* }}} */

/* {{{ _pop3_fetch_message_part() */

static void _pop3_fetch_message_part(INTERNAL_FUNCTION_PARAMETERS, int mode)
{
	POP3_DECLARE_INIT_OBJECT (object)
	zval	*session_container;
	char	*pop3_return_val;
	int	*socket_id;
	char	*message;
	long	message_id, n_lines = 0;

	/* Retrieve session data and other parameters */
	if (object) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l|l", &message_id, &n_lines) == FAILURE) {
			RETURN_FALSE;
		}
		POP3_FROM_OBJECT(socket_id, object);
	} else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl|l", &session_container, &message_id, &n_lines) == FAILURE) {
			RETURN_FALSE;
		}
		POP3_FETCH_SESSION(socket_id, session_container);
	}
	
	switch (mode) {
		case PHP_POP3_HEADER:
			/* Retrieve message headers */
			POP3_CALL(pop3_top(*socket_id, message_id, 0))
			break;
		case PHP_POP3_MESSAGE:
			if (!n_lines) { /* Retrieve message */
				POP3_CALL(pop3_retr(*socket_id, message_id))
			} else { /* Retrieve n_lines from the message */
				POP3_CALL(pop3_top(*socket_id, message_id, n_lines))
			}
			break;
		case PHP_POP3_DELETE:
			/* Delete message */
			POP3_CALL(pop3_dele(*socket_id, message_id))
			RETURN_TRUE;
			break;
		default: /* can't happen */
			RETURN_FALSE;
	}

	message = retr2msg(pop3_return_val);
	free(pop3_return_val);

	POP3_SAFE_POINTER(message);

	RETVAL_STRING(message, 1);

	free(message);
}

/* }}} */

/* {{{ proto string pop3_get_message_header (resource session, int message_id)
  */
PHP_FUNCTION (pop3_get_message_header)
{
	_pop3_fetch_message_part(INTERNAL_FUNCTION_PARAM_PASSTHRU, PHP_POP3_HEADER);	
}
/* }}} */

/* {{{ proto string pop3_get_message (resource session, int message_id, [int n_lines])
  */
PHP_FUNCTION (pop3_get_message)
{
	_pop3_fetch_message_part(INTERNAL_FUNCTION_PARAM_PASSTHRU, PHP_POP3_MESSAGE);
}
/* }}} */

/* {{{ proto bool pop3_delete_message (resource session, int message_id)
  */
PHP_FUNCTION (pop3_delete_message)
{
	_pop3_fetch_message_part(INTERNAL_FUNCTION_PARAM_PASSTHRU, PHP_POP3_DELETE);
}
/* }}} */

/* {{{ proto bool pop3_undelete (resource session)
  */
PHP_FUNCTION (pop3_undelete)
{
	POP3_DECLARE_INIT_OBJECT (object)
	zval	*session_container;
	int	*socket_id;
	char    *pop3_return_val;

	if (object) {
		if (ZEND_NUM_ARGS()) {
			WRONG_PARAM_COUNT;
		}
		POP3_FROM_OBJECT(socket_id, object);
	} else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &session_container) == FAILURE) {
			RETURN_FALSE;
		}
		POP3_FETCH_SESSION(socket_id, session_container);
	}

	POP3_CALL(pop3_rset(*socket_id))

	RETURN_TRUE;
}
/* }}} */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
